from dateutil.parser import parse
from .utils import is_four_digit_year
from .clean_value import number_cleaner, string_cleaner, date_from_timestamp

import pandas as pd

def clean_dates(values):
    test_values = values[:100]
    if is_four_digit_year(test_values):
        return pd.to_datetime(values,format='%Y',errors='coerce')

    utc = False
    is_int = False
    for val in test_values:
        if not val:
            continue
        try:
            d = val
            if type(d) != datetime:
                d = parse(unicode(val))
            if d.tzinfo is not None and d.tzinfo.utcoffset(d) is not None:
                utc = True
        except:
            try:
                d = date_from_timestamp(val)
                if d.year > 1900 and d.year < datetime.now().year + 5:
                    is_int = True
            except:
                continue

    if is_int:
        return [date_from_timestamp(d) for d in values]
    
    return pd.to_datetime(values, errors='coerce', utc=utc).values


def clean_coordinates(values):
    series = pd.Series(values).str.replace(')','')
    series = series.str.replace('(','')
    return series.values


def clean_numbers(values):
    series = pd.Series(values).apply(number_cleaner)
    return series.values



import urllib
from urllib.parse import urlparse
import os
import time
import re

def is_four_digit_year(values):
    new_values = [v for v in values if v and str(v) != 'nan']
    if float(len(new_values))/len(values) > .5 and all([len(str(v)) is 4 and str(v)[0] in ['1','2'] for v in new_values]):
        return True

    return False



import pandas as pd
import numpy as np
from .definitions import NUMBERS

interval_to_pd = {
    'year': 'A',
    'month': 'M',
    'week': 'W',
    'day': 'D',
    'hour': 'H',
    'minute': 'T',
    'second': 'S'
}

def best_intervals(start_date, end_date):
    try:
        delta = pd.to_datetime(end_date, coerce=True) - pd.to_datetime(start_date, coerce=True)
        days = delta.days
    except:
        return ['year']
    
    intervals = []
    if days > 356 * 3:
        intervals.append('year')
    if days > 90 and days < 365 * 20:
        intervals.append('month')

    if days > 21 and days < 365 * 5:
        intervals.append('week')
        
    if days > 3 and days < 365 * 2:
        intervals.append('day')
    
    if days <=3:
        seconds = delta.seconds
        if seconds // 3600 >= 3:
            intervals.append('hour')
        if seconds // 60 % 60 >= 3 and seconds // 60 % 60 < 300:
            intervals.append('minute')
        if seconds < 300:
            intervals.append('second')

    return intervals


def agg_by_date(df, datetime_header, number_headers, interval=None, agg='mean'):
    if agg not in ['mean', 'sum']:
        return None

    df = df[pd.notnull(df[datetime_header])]
    
    dates = pd.to_datetime(df[datetime_header],infer_datetime_format=True)
    indexed = df.set_index(pd.DatetimeIndex(dates))
    series = indexed[number_headers]

    last = df[datetime_header].max()
    first = df[datetime_header].min()
    
    if interval:
        has_interval = True
    else:
        has_interval = False
        interval = best_intervals(first,last)
        if len(interval) > 0:
            interval = interval[0]

    grouped = series.groupby(pd.TimeGrouper(freq=interval_to_pd[interval]))
    df_grouped_by_date = getattr(grouped,agg)().reset_index()

    x_values = df_grouped_by_date[datetime_header].dt.strftime('%Y-%m-%d %H:%M:%S').values
    y_values = []
    for number_header in number_headers:
        y_values.append(df_grouped_by_date[number_header].values)

    output = np.array([
        x_values,
        y_values,
    ],dtype='O')
    
    # output results
    return output


def agg_by_category_by_date(df, datetime_header, number_headers, category_headers, 
                                    interval=None, agg='mean'):
    

    all_column_headers = category_headers + number_headers

    df = df[pd.notnull(df[datetime_header])]

    indexed = df.set_index(pd.DatetimeIndex(df[datetime_header]))
    
    last = df[datetime_header].max()
    first = df[datetime_header].min()

    if interval:
        has_interval = True
    else:
        has_interval = False
        interval = best_intervals(first,last)
        if len(interval) > 0:
            interval = interval[0]

    grouper = pd.TimeGrouper(freq=interval_to_pd[interval])
    grouped = indexed[all_column_headers].groupby([grouper] + category_headers)
    df_grouped_by_date = getattr(grouped,agg)().reset_index()

    x_values = df_grouped_by_date[datetime_header].dt.strftime('%Y-%m-%d %H:%M:%S').values

    y_values = []
    z_values = []
    for number_header in number_headers:
        y_values.append(df_grouped_by_date[number_header].values)

    for category_header in category_headers:
        z_values.append(df_grouped_by_date[category_header].values)

    output = np.array([
        x_values,
        y_values,
        z_values
    ],dtype='O')
    
    # output results
    return output



import pandas as pd
from typing import List, NamedTuple

from .timeseries import agg_by_category_by_date
from primitive_interfaces.base import PrimitiveBase

class AggregateByDateTimeCategory(PrimitiveBase[pd.DataFrame, List[str]]):
    __author__ = 'distil'

    def __init__(self):
        pass

    def get_params(self) -> dict:
        return {}

    def set_params(self, params: dict) -> None:
        self.params = params

    def get_call_metadata(self) -> {}: 
        return {}

    def fit(self):
        pass

    def produce(self, inputs: pd.DataFrame, values: List[str] = [], groupby : List[str] = [],
                datetime=None, intervals=None, aggregation=None):
        return agg_by_category_by_date(inputs, datetime, values, groupby, interval=intervals, 
                            agg=aggregation)



import numpy as np
from sklearn.decomposition import PCA
from typing import NamedTuple, List
from primitive_interfaces.base import PrimitiveBase

Input = np.ndarray
Output = List[int]
Params = dict
CallMetadata = dict

class PCAFeatures(PrimitiveBase[Input, Output, Params]):
    __author__ = "distil"
    __metadata__ = {
        "id": "f7543334-97f7-3c43-978d-2d0b03829d4d",
        "name": "punk.feature_selection.pca.PCAFeatures",
        "common_name": "PCAFeatures",
        "description": "Ranking of features using principal component analysis. Returns a ranking of the features based on the magnitude of their contributions to the first principal componenet and a ranking of the features based on the highest magnitude contribution to all the principal componenets.",
        "languages": [
            "python3.6"
        ],
        "library": "punk",
        "version": "1.1.0",
        "source_code": "https://github.com/NewKnowledge/punk/blob/dev/punk/feature_selection/pca.py",
        "algorithm_type": [                                                         
            "dimensionality reduction"                                              
        ],
        "task_type": [
            "feature extraction"
        ],
        "output_type": [
            "features"
        ], 
        "team": "distil",
        "interface_type" : "data_cleaning",
        "schema_version": 1.0,
        "build": [
            {
                "type": "pip",
                "package": "punk"
            }
        ],
         "compute_resources": {
            "sample_size": [
                1000.0, 
                10.0
            ],
            "sample_unit": [
                "MB"
            ],
            "num_nodes": [
                1
            ],
            "cores_per_node": [
                1
            ],
            "gpus_per_node": [
                0
            ],
            "mem_per_node": [
                1.0
            ],
            "disk_per_node": [
                1.0
            ],
            "mem_per_gpu": [
                0.0
            ],
            "expected_running_time": [
                5.0
            ]
        }
    }


    def __init__(self) -> None:
        self.callMetadata = {}
        self.params = {}
        pass

    def fit(self) -> None:
        pass

    def get_params(self) -> Params:
        return self.params

    def set_params(self, params: Params) -> None:
        self.params = params

    def get_call_metadata(self) -> CallMetadata:
        return self.callMetadata

    def produce(self, inputs: Input) -> Output:
        """ Perform PCA and return a list of the indices of the most important
        features, ordered by contribution to first PCA component
                                                                                
        The matrix M will correspond to the absolute value of the components of 
        the decomposiiton thus giving a matrix where each column corresponds to 
        a principal component and rows are the components that rescale each 
        feature.  
                                                                                
        For example, pca.components_.T could look like this:                        
                                                                                
        array([[ 0.52237162,  0.37231836, -0.72101681, -0.26199559],                
            [-0.26335492,  0.92555649,  0.24203288,  0.12413481],                 
            [ 0.58125401,  0.02109478,  0.14089226,  0.80115427],                 
            [ 0.56561105,  0.06541577,  0.6338014 , -0.52354627]])                
                                                                                
        So the most important feature with respect to the first principal 
        component would be the 3rd feature as this has an absolute value of 
        0.58125401 which is greater than all the entries in that column.                             
                                                                                
                                                                                
        "importance_on1stpc" corresponds to the indices of most important       
        features for the first principal. Component are in ascending order      
        (most important feature 0, least important feature n_features-1).       
                                                                                
        
        Params 
        ------- 
        data : np.ndarray, [n_samples, n_features]
            Training data.
        """

        pca = PCA()

        try:
            pca.fit_transform(inputs)

            M = np.absolute(pca.components_.T)

            # Rank features based on contribtuions to 1st PC
            self.importance_on1stpc = np.argsort(M[:,0], axis=0)[::-1]
        except:
            # If any error occurs, just return indices in original order
            # In the future we should consider a better error handling strategy
            self.importance_on1stpc = [i for i in range(inputs.shape[0])]

        return self.importance_on1stpc


import numpy as np
from typing import Tuple
from sklearn.decomposition import PCA, FactorAnalysis
from sklearn.model_selection import cross_val_score
from primitive_interfaces.base import PrimitiveBase

Inputs = np.ndarray
Outputs = Tuple[Tuple[float, int], Tuple[float, int]]
Params = dict
CallMetadata = dict

class HeteroscedasticityTest(PrimitiveBase[Inputs, Outputs, Params]):
    __author__ = 'distil'
    __metadata__ = {
        "id": "25a459ea-a219-37ba-a471-b2e6a5ddd5b8",
        "name": "punk.novelty_detection.heteroscedasticity.HeteroscedasticityTest",
        "common_name": "HeteroscedasticityTest",
        "description": "Test heteroscedaticity and optimal principal subspace of data.",
        "languages": [
            "python3.6"
        ],
        "library": "punk",
        "version": "1.1.0",
        "source_code": "https://github.com/NewKnowledge/punk/blob/dev/punk/novelty_detection/heteroscedasticity.py",

        "algorithm_type": [                                                         
            "dimensionality reduction"                                              
        ],
        "task_type": [
            "feature extraction"
        ],
        "output_type": [
            "features"
        ], 
        "team": "distil",
        "interface_type" : "data_cleaning",
        "schema_version": 1.0,
        "build": [
            {
                "type": "pip",
                "package": "punk"
            }
        ],
        "compute_resources": {
            "sample_size": [
                1000.0, 
                10.0
            ],
            "sample_unit": [
                "MB"
            ],
            "num_nodes": [
                1
            ],
            "cores_per_node": [
                1
            ],
            "gpus_per_node": [
                0
            ],
            "mem_per_node": [
                1.0
            ],
            "disk_per_node": [
                1.0
            ],
            "mem_per_gpu": [
                0.0
            ],
            "expected_running_time": [
                5.0
            ]
        }
    }

    __test__ = False

    def __init__(self, max_iter: int = 1000000, tol: int = 1e-8):
        self.max_iter = max_iter
        self.tol      = tol
        self.params = {'max_iter': self.max_iter, 'tol': self.tol}
        self.callMetadata = {}

    def fit(self) -> None:
        pass

    def get_params(self) -> Params:
        return self.params

    def set_params(self, params: Params) -> None:
        self.params = params

    def get_call_metadata(self) -> CallMetadata:
        return self.callMetadata

    def produce(self, inputs: Inputs) -> Outputs:
        """ Test heteroscedaticity of your data.

        The consequence is that the likelihood of new data can be used for 
        model selection and covariance estimation.

        Code taken from 'http://scikit-learn.org/stable/auto_examples/'             
        'decomposition/plot_pca_vs_fa_model_selection.html'.                        

        Params
        ------  
        data : array-like
            Training data.

        Returns
        -------
        "pca" returns a two tuple of number of componenets and likelihood for 
        best principal componenets estimator.                                   
        "fa" returns a two tuple of number of componenets and likelihood for 
        best factor analysis estimator.
        """
        pca = PCA(svd_solver='full')                                                
        fa = FactorAnalysis(max_iter=self.max_iter, tol=self.tol)                             

        X = inputs

        pca_scores, fa_scores = [], []                                              
        n_featues = X.shape[1]                                                      
        if n_featues > 20:                                                          
            n_components = np.arange(0, n_featues+1, 5)                             
        else:                                                                       
            n_components = np.arange(0, n_featues+1)                                
                                                                                
        for n in n_components:                                                      
            pca.n_components = n                                                    
            fa.n_components = n                                                     
            pca_scores.append( (np.mean(cross_val_score(pca, X)), n) )              
            fa_scores.append( (np.mean(cross_val_score(fa, X)), n) )                

        self.pca = max(pca_scores, key=lambda s: s[0])
        self.fa  = max(fa_scores, key=lambda s: s[0])

        return (self.pca, self.fa)


import pandas as pd
import numpy as np

def agg_by_categories(df, category_headers, number_headers, agg):

    all_headers = number_headers + category_headers

    temp_grouped = df[all_headers].groupby(category_headers)
    grouped = getattr(temp_grouped,agg)().reset_index()

    categories = {}
    numbers = {}

    for category_header in category_headers:
        categories[category_header] = grouped[category_header].values

    for number_header in number_headers:
        numbers[number_header] = grouped[number_header].values

    output = np.array([
        categories,
        numbers
    ],dtype='O')
    
    # output results
    return output


import pandas as pd
from sklearn.preprocessing import MinMaxScaler
import numpy as np
import random


def encode_df(df, data_types={}):
    """
    Encode columns so their values are usable in vector operations. Making a few 
    assumptions here, like that datetime should be an integer, and that it's 
    acceptable to fill NaNs with 0.
    """

    numbers = []
    categories = []
    datetimes = []
        
    for column, series in df.iteritems():
        if column in data_types:
            dtypes = data_types[column]
        else:
            dtypes = {}
        
        if 'datetime' in dtypes:
            index = pd.DatetimeIndex(pd.to_datetime(series.values))
            df[column] = index.astype(np.int64).astype(float)
            datetimes.append(column)
            
        elif any([dtype in dtypes for dtype in ['category', 'label']]):
            # If this can be treated as a category, encode it
            df = pd.get_dummies(df,columns=[column])
            categories.append(column)
                    
        elif 'numeric' in dtypes:
            df[column] = clean_numbers(df[column].values)
            df[column] = df[column].astype(float)
            numbers.append(column)
            
        else:
            df.drop(column,1,inplace=True)

    # Scale continuous columns
    scaler = MinMaxScaler()
    df[numbers + datetimes] = scaler.fit_transform(df[numbers + datetimes])
    
    return df, categories, numbers, datetimes




import numpy as np
import pandas as pd
from datetime import datetime

def strip_non_ascii(string):
    ''' Returns the string without non ASCII characters'''
    if type(string) == datetime:
        return str(string)
    stripped = (c for c in string if 0 < ord(c) < 127)
    return ''.join(stripped)


def date_from_timestamp(d):
    try:
        return datetime.fromtimestamp(int(d))
    except:
        return None


def header_cleaner(string):
    header = strip_non_ascii(string)
    header = header.replace('\n','').replace('\t','').replace('\r','')
    return header


def number_cleaner(x,force=False):
    try:
        x_str = unicode(x).strip()
    except:
        return np.nan

    if x_str == unicode('nan'):
        if force:
            return 0.0
        else:
            return np.nan

    x_str = ''.join([c for c in unicode(x) if c == '.' or c.isdigit() or c == '-' or c == 'e'])
    if not x_str:
        if force:
            return 0.0
        else:
            return np.nan

    try:
        return float(x_str) 
    except:
        return np.nan


def string_cleaner(x):
    if isinstance(x, unicode):
        return x
    if isinstance(x, basestring):
        return unicode(x, errors='ignore')
    else:
        return unicode(str(x), errors='ignore')



import pandas as pd
import numpy as np

def range_groups(df, number_headers, bins=None):
    max_bins = 20

    df_desc = df[number_headers].describe().reset_index()
    df_nums = df[number_headers]
    df_nums = df_nums.dropna()

    if not bins:
        lowest_min_header = None
        lowest_min_value = None
        highest_max_header = None
        highest_max_value = None

        for number_header in number_headers:
            min_val = df_desc.loc[df_desc['index'] == 'min'][number_header].values[0]
            max_val = df_desc.loc[df_desc['index'] == 'max'][number_header].values[0]

            if not lowest_min_value or min_val < lowest_min_value:
                lowest_min_header = number_header
                lowest_min_value = min_val

            if not highest_max_value or max_val > highest_max_value:
                highest_max_header = number_header
                highest_max_value = max_val

        high_low = np.concatenate([df_nums[lowest_min_header].values,df_nums[highest_max_header].values])
        high_low = high_low[~np.isnan(high_low)]
        counts,bins = np.histogram(high_low,bins='auto')

        if len(counts) > max_bins:
            bins = max_bins

    ys = {}
    x_values = None
    for number_header in number_headers:
        count,division = np.histogram(df_nums[number_header].values,bins=bins)

        if not x_values:
            x_values = []
            for i,d in enumerate(division):
                if i == len(division) - 1:
                    break

                x_values.append('{0:.2f}'.format(d) + u' to ' + '{0:.2f}'.format(division[i+1]))

        y_values = list(count)
        ys[number_header] = y_values

    output = np.array([
        x_values,
        y_values
    ],dtype='O')
    
    return output



from setuptools import setup, find_packages


setup(
    name="punk",

    version="1.1.0",

    description="Primitives for Uncovering New Knowledge.",
    long_description="Machine Learning pipeline elements.",

    url="https://github.com/NewKnowledge/punk",

    author="New Knowledge",
    author_email="support@newknowledge.io",

    license="MIT",

    classifiers=[
        "Development Status :: 4 - Beta",

        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",

        # Pick your license as you wish (should match "license" above)
        "License :: OSI Approved :: MIT License",

        "Programming Language :: Python :: 3.6",

        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],

    keywords="TA1 primitive, feature selection, novelty detection",


    packages=find_packages(exclude=['tests']),

    install_requires=["numpy", "scikit-learn", "scipy", "python-dateutil"],

    entry_points = {
        'd3m.primitives': [
            'distil.PCAFeatures = punk.feature_selection.pca:PCAFeatures',
            'distil.RFFeatures = punk.feature_selection.rf:RFFeatures',
            'distil.Heteroscedasticity = punk.novelty_detection.heteroscedasticity:HeteroscedasticityTest',
            'distil.AggregateByCategory = punk.aggregator.aggregateByCategory:AggregateByCategory',
            'distil.AggregateByDateTime = punk.aggregator.aggregateByDateTime:AggregateByDateTime',
            'distil.AggregateByNumericRange = punk.aggregator.aggregateByNumericRange:AggregateByNumericRange',
            'distil.CleanDates = punk.preppy.cleanDates:CleanDates',
            'distil.CleanNumbers = punk.preppy.cleanNumbers:CleanNumbers',
            'distil.CleanStrings = punk.preppy.cleanStrings:CleanStrings'
        ],
    },

)




# Liberally borrowed http://stackoverflow.com/a/21515453/2367526

def split_obj (obj, prefix = None):
    '''
    Split the object, returning a 3-tuple with the flat object, optionally
    followed by the key for the subobjects and a list of those subobjects.
    '''
    # copy the object, optionally add the prefix before each key

    new = obj.copy() if prefix is None else { '{}_{}'.format(prefix, k): v for k, v in obj.items() }

    # try to find the key holding the subobject or a list of subobjects
    for k, v in new.items():
        # list of subobjects
        if isinstance(v, list):
            # this is a list of strings or ints, which is a one-to-many
            # can't deal with that here, but leave the data in case it's 
            # useful downstream
            if not isinstance(v[0], dict):
                new[k] = ','.join(v)
                return new, None, None
            del new[k]
            return new, k, v
        # or just one subobject
        elif isinstance(v, dict):
            del new[k]
            return new, k, [v]
    return new, None, None


def flatten_json(data, prefix = None):
    '''
    Flatten the data, optionally with each key prefixed.
    '''
    # iterate all items
    for item in data:
        # split the object
        flat, key, subs = split_obj(item, prefix)

        # just return fully flat objects
        if key is None:
            yield flat
            continue

        # otherwise recursively flatten the subobjects
        for sub in flatten_json(subs, key):
            sub.update(flat)
            yield sub



import pandas as pd
from typing import List

from .clean_value import string_cleaner
from primitive_interfaces.base import PrimitiveBase

Inputs = pd.DataFrame
Outputs = List[float]
Params = dict
CallMetadata = dict

class CleanStrings(PrimitiveBase[Inputs, Outputs, Params]):
    __author__ = 'distil'
    __metadata__ = {
        "id": "447a0205-1066-3a6d-a001-e4822a0e6ed7",
        "name": "punk.preppy.cleanStrings.CleanStrings",
        "common_name": "CleanDates",
        "description": "Forcing just about anything that looks like a string into a unicode object",
        "languages": [
            "python3.6"
        ],
        "library": "punk",
        "version": "1.1.0",
        "source_code": "https://github.com/NewKnowledge/punk/blob/dev/punk/preppy/cleanStrings.py",
        "is_class": True,
        "interface_type": "data_cleaning",
        "algorithm_type": [                                                         
            "data_preparation"                                              
        ],
        "task_type": [
            "data cleaning"                                              
        ],
        "output_type": [
            "features"
        ], 
        "team": "distil",
        "schema_version": 1.0,
        "build": [
            {
                "type": "pip",
                "package": "punk"
            }
        ],
        "compute_resources": {
            "sample_size": [
                1000.0, 
                10.0
            ],
            "sample_unit": [
                "MB"
            ],
            "num_nodes": [
                1
            ],
            "cores_per_node": [
                1
            ],
            "gpus_per_node": [
                0
            ],
            "mem_per_node": [
                1.0
            ],
            "disk_per_node": [
                1.0
            ],
            "mem_per_gpu": [
                0.0
            ],
            "expected_running_time": [
                5.0
            ]
        }
    }


    def __init__(self):
        pass

    def get_params(self) -> Params:
        return {}

    def set_params(self, params: Params) -> None:
        self.params = params

    def get_call_metadata(self) -> CallMetadata: 
        return {}

    def fit(self) -> None:
        pass

    def transform(self, inputs: Inputs) -> Outputs:
        output = []
        for l in inputs:
            output.append([string_cleaner(s) for s in l])

        return output